/*  FMALLBUG.C
 *  this program reveals a problem with Microsoft's
 *  far memory allocation function, _fmalloc()
 *  last mod.: 2-JUN-91
 */

#include <STDIO.H>
#include <STDLIB.H>
#include <FCNTL.H>
#include <CONIO.H>
#include <IO.H>
#include <MALLOC.H>     /*  for compilation only with Microsoft C  */
#include <STRING.H>

#include <L_MEM.H>

typedef struct mb
    {
    struct mb far *mb_previous;
    int mb_data;
    } Mem_block;

Mem_block far *list;
Mem_block far *p;
Uint first_alloc_size =  40000;
Uint second_alloc_size = 46000;
Ulong first_allocation, second_allocation;
int err_flag, ch;
char msg[81];

void main(int argc, char **argv);
void display_first_allocation(void);
void display_second_allocation(void);
Ulong msc_allocation(Uint block_size);
Ulong dolphin_allocation(Uint block_size, int *err_flag);
void compact_mem_log(void);
void print_reports(void);

/*---------------------------*/
void main(int argc,char **argv)
{
set_mem_alloc_strategy(atoi(argv[1]));

printf("\n\nmem_alloc() without memory log (max. avail.: %lu bytes):",
    mem_max(&err_flag));
first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
display_first_allocation();
second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
display_second_allocation();

set_mem_log(TRUE);
/*  write program name to log  */
strcpy(msg,"Program is ");
strcat(msg,argv[0]);
_L_(mem_log_msg(msg,&err_flag);)

printf("\n\nmem_alloc() with memory log (max. avail.: %lu bytes):",
    mem_max(&err_flag));
first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
display_first_allocation();
second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
display_second_allocation();
printf("\n");

compact_mem_log();
sprintf(msg,"mem_alloc() with memory log and compaction "
    "(max. avail.: %lu bytes)",mem_max(&err_flag));
printf("\n%s:",msg);
/*  put the message in the memory log  */
_L_(mem_log_msg(msg,&err_flag);)
first_allocation = dolphin_allocation(first_alloc_size,&err_flag);
display_first_allocation();
compact_mem_log();
second_allocation = dolphin_allocation(second_alloc_size,&err_flag);
display_second_allocation();
printf("\n");

/*  can't do this at the beginning because _fmalloc() ties up memory
 *  (as this program demonstrates) and so mem_alloc() does not succeed
 */
compact_mem_log();
printf("\nUsing _fmalloc() (max. available: %lu bytes):",
    mem_max(&err_flag));
first_allocation = msc_allocation(first_alloc_size);
display_first_allocation();
second_allocation = msc_allocation(second_alloc_size);
display_second_allocation();
print_reports();
}

/*-------------------------------*/
void display_first_allocation(void)
{
Uint num_blocks = (Uint)(first_allocation/first_alloc_size);

printf("\nUsing block size %u, %lu bytes allocated (%u block%s)%s.",
    first_alloc_size,first_allocation,num_blocks,
    ( num_blocks==1 ? "" : "s" ),
    ( num_blocks==0 ? "" : " and freed" ) );
}

/*--------------------------------*/
void display_second_allocation(void)
{
Uint num_blocks = (Uint)(second_allocation/second_alloc_size);

printf("\nUsing block size %u, %lu bytes allocated (%u block%s)%s.",
    second_alloc_size,second_allocation,num_blocks,
    ( num_blocks==1 ? "" : "s" ),
    ( num_blocks==0 ? "" : " and freed" ) );
}

/*---------------------------------*/
Ulong msc_allocation(Uint block_size)
{
Ulong total_allocation = 0L;

list = (Mem_block far *)NULL;

/*    allocate memory blocks until memory exhausted  */
while ( ( p = (Mem_block far *)_fmalloc(block_size) )
    != (Mem_block far *)NULL )
    {
    p->mb_previous = list;
    list = p;
    total_allocation += block_size;
    }

/*    free the allocated blocks  */
while ( list != (Mem_block far *)NULL )
    {
    p = list;
    list = list->mb_previous;
    _ffree(p);
    }

return ( total_allocation );
}

/*---------------------------------------------------*/
Ulong dolphin_allocation(Uint block_size,int *err_flag)
{
Ulong total_allocation = 0L;

list = (Mem_block far *)NULL;
*err_flag = NO_ERROR;

/*    allocate memory blocks until memory exhausted  */
while ( TRUE )
    {
    _L_(p = (Mem_block far *)mem_alloc((Ulong)block_size,err_flag);)
    if ( *err_flag )
        break;
    p->mb_previous = list;
    list = p;
    total_allocation += block_size;
    }

if ( *err_flag == INSUFFICIENT_MEMORY )
    *err_flag = NO_ERROR;

if ( *err_flag )
    {
    printf("\n%s",mem_err_msg(*err_flag,msg));
    exit(1);
    }
else
    {
    /*    free the allocated blocks  */
    while ( TRUE )
        {
        p = list;
        list = list->mb_previous;
        _L_(mem_free((Far_str_ptr)p,err_flag);)
        if ( *err_flag == NULL_POINTER )
            break;
        else if ( *err_flag != NO_ERROR )
            {
            printf("\n%s",mem_err_msg(*err_flag,msg));
            exit(2);
            }
        }
    }

return ( total_allocation );
}

/*----------------------*/
void compact_mem_log(void)
{
_L_(compact_log(&err_flag);)
if ( err_flag )
    {
    printf("\nError when compacting log: %s",mem_err_msg(err_flag,msg));
    mem_log_msg("Error when compacting log.",&err_flag);
    }
else
    {
    printf("\nMemory log compacted.");
    _L_(mem_log_msg("Memory log compacted.",&err_flag);)
    }
}

/*--------------------*/
void print_reports(void)
{
FILE *output;
char *logfile = "FMALLBUG.LOG";

printf("\n\nMemory allocation log has %u entries.",mem_log_entries());
printf("\nPress P to send reports #1 and #2 to printer,"
       "\nF to send reports to disk file %s,"
       "\nany other key to quit ... ",logfile);

while ( kbhit() )
    getch();
ch = getche();
if ( ch == 'P' || ch == 'p' )
    {
    setmode(fileno(stdprn),O_TEXT);
    mem_log_report(1,stdprn,1,TRUE,&err_flag);
    mem_log_report(2,stdprn,1,TRUE,&err_flag);
    }
else if ( ch == 'F' || ch == 'f' )
    {
    if ( ( output = fopen(logfile,"wt") ) == NULL )
        {
        printf("\nUnable to open file %s",logfile);
        exit(1);
        }
    mem_log_report(1,output,1,TRUE,&err_flag);
    if ( err_flag )
        printf("\nError when writing first report: %s",
            mem_err_msg(err_flag,msg));
    else
        {
        mem_log_report(2,output,1,TRUE,&err_flag);
        if ( err_flag )
            printf("\nError when writing second report: %s",
            mem_err_msg(err_flag,msg));
        }
    }

printf("\n");
}
